#!/usr/bin/env python

"""
Generate RSVP code from simplified nodes.

Copyright (C) 2007 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from simplify.simplified import *
import rsvp

# Classes.

class Generator(Visitor):

    "An RSVP code generator."

    def __init__(self):

        "Initialise the visitor."

        Visitor.__init__(self)
        self.assemblers = []
        self.current_assemblers = []

        # Satisfy visitor issues.

        self.visitor = self

    def get_code(self):
        return rsvp.get_image(*self.assemblers)

    def assembler(self):
        return self.current_assemblers[-1]

    def process(self, module):

        """
        Process the given 'module', creating RSVP assemblers to store code and
        data.
        """

        self.module = module

        main = rsvp.RSVPAssembler(1)
        self.current_assemblers.append(main)

        # Store constants.

        for name, const in module.simplifier.constants.items():
            main.label("const$%s$%s" % (self.module.name, name))
            main.add(const.value)

        # Reserve the namespace.

        for name in module.namespace.keys():
            main.label("global$%s$%s" % (self.module.name, name))
            main.add(None)

        # Generate subprograms.

        for subprogram in module.simplifier.subprograms:
            if not subprogram.internal:

                # Generate specific subprograms.

                for specialisation in subprogram.specialisations():
                    self.process_subprogram(specialisation)

        # Generate the initialisation.

        self.dispatch(module)

        # Remember the assembler used.

        self.current_assemblers.pop()
        self.assemblers.append(main)

    def default(self, node, *args):
        if hasattr(node, "code"):
            self.dispatches(node.code)

    def visitInvoke(self, invoke):

        "Process the given 'invoke' node."

        sub = self.assembler()

        # NOTE: Generate arguments in a way compatible with subprogram
        # NOTE: consumption as parameters.

        for subprogram in invoke.invocations:
            consumed_args = invoke.consumed_args[subprogram]

            # NOTE: Generate switch table.

            # Save the current frame.

            sub.add("SCF")

            # Extend the stack for the arguments.

            nparams = len(consumed_args)
            if not invoke.share_locals:
                sub.add("ESF", nparams)

                # For each argument, evaluate the expression.

                for arg in invoke.args:
                    if arg is None:
                        continue

                    self.dispatch(arg)

                    # Then, save the argument to the correct part of the arguments
                    # frame.

                    pos = self.arg_position(consumed_args, arg)
                    sub.add("MVA", pos)

                # Reserve the namespace.

                sub.add("NSF", nparams)

            sub.add("JAS", "sub$%s$%s" % (subprogram.module.name, subprogram.full_name()))

            if not invoke.share_locals:
                sub.add("PSF") # previous stack frame

    visitInvokeFunction = visitInvoke
    visitInvokeRef = visitInvoke

    def visitModule(self, module):

        "Process the given 'module'."

        self.dispatches(module.code)

    def visitSubprogram(self, subprogram):

        "Process the 'subprogram'."

        if subprogram.internal:
            self.process_subprogram(subprogram)

    def process_subprogram(self, subprogram):

        sub = rsvp.RSVPAssembler(1)
        self.current_assemblers.append(sub)

        sub.label("sub$%s$%s" % (self.module.name, subprogram.full_name()))

        locals = subprogram.namespace.keys()
        params = subprogram.paramtypes.keys()
        nparams = len(params)
        non_params = len(locals) - nparams

        if non_params > 0:
            sub.add("ESF", non_params) # extend stack frame for locals

        # Produce the standard end of subprogram return.

        sub.add("RAC") # return

        self.current_assemblers.pop()
        self.assemblers.append(sub)

    def arg_position(self, consumed_args, arg):

        """
        Return, for the specified 'consumed_args', the position of 'arg' in the
        list, handling keyword arguments appropriately.
        """

        if isinstance(arg, Keyword):
            arg = arg.expr
        return consumed_args.index(arg)

# Convenience functions.

def generate(module):
    generator = Generator()
    generator.process(module)
    return generator.get_code()

# vim: tabstop=4 expandtab shiftwidth=4
